use std::error::Error;
use std::sync::Arc;
use vulkano::buffer::Subbuffer;
use vulkano::command_buffer::{AutoCommandBufferBuilder, PrimaryAutoCommandBuffer};
use vulkano::device::Device;
use vulkano::memory::allocator::StandardMemoryAllocator;
use vulkano::pipeline::GraphicsPipeline;
use vulkano::swapchain::Swapchain;

use crate::renderer::{create_triangle_pipeline, Vertex2D};

const VERTICES: [Vertex2D; 12] = [
    // Triangle en haut à gauche
    Vertex2D {
        position: [-0.5, -0.75],
        color: [1.0, 0.0, 0.0],
    },
    Vertex2D {
        position: [-0.75, -0.25],
        color: [0.0, 1.0, 0.0],
    },
    Vertex2D {
        position: [-0.25, -0.25],
        color: [0.0, 0.0, 1.0],
    },
    // Triangle en bas à gauche
    Vertex2D {
        position: [-0.5, 0.25],
        color: [0.5, 0.5, 0.5],
    },
    Vertex2D {
        position: [-0.75, 0.75],
        color: [0.2, 0.8, 0.2],
    },
    Vertex2D {
        position: [-0.25, 0.75],
        color: [0.8, 0.2, 0.2],
    },
    // Triangle en haut à droite
    Vertex2D {
        position: [0.5, -0.75],
        color: [1.0, 1.0, 0.0],
    },
    Vertex2D {
        position: [0.25, -0.25],
        color: [0.0, 1.0, 1.0],
    },
    Vertex2D {
        position: [0.75, -0.25],
        color: [1.0, 0.0, 1.0],
    },
    // Triangle en bas à droite
    Vertex2D {
        position: [0.5, 0.25],
        color: [0.1, 0.5, 0.8],
    },
    Vertex2D {
        position: [0.25, 0.75],
        color: [0.8, 0.6, 0.1],
    },
    Vertex2D {
        position: [0.75, 0.75],
        color: [0.3, 0.4, 0.6],
    },
];

pub struct Scene {
    pipeline: Arc<GraphicsPipeline>,
    vertex_buffer: Subbuffer<[Vertex2D]>,
}

impl Scene {
    pub fn load(
        device: &Arc<Device>,
        swapchain: &Arc<Swapchain>,
        memory_allocator: &Arc<StandardMemoryAllocator>,
    ) -> Result<Self, Box<dyn Error>> {
        let pipeline = create_triangle_pipeline(device, swapchain)?;
        let vertex_buffer = Vertex2D::create_buffer(Vec::from_iter(VERTICES), memory_allocator)?;

        Ok(Scene {
            pipeline,
            vertex_buffer,
        })
    }

    pub fn render(
        &self,
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
    ) -> Result<(), Box<dyn Error>> {
        let vertex_count = self.vertex_buffer.len() as u32;
        let instance_count = vertex_count / 3;

        unsafe {
            builder
                .bind_pipeline_graphics(self.pipeline.clone())?
                .bind_vertex_buffers(0, self.vertex_buffer.clone())?
                .draw(vertex_count, instance_count, 0, 0)?;
        }

        Ok(())
    }
}
